/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * This file is part of SableCC.                             *
 * See the file "LICENSE" for copyright information and the  *
 * terms and conditions for copying, distribution and        *
 * modification of SableCC.                                  *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/*
* Last Modification date : October, the 11th 2003
* Goal of the modification : addition of specifier to
* the generated production($prod) to handle list of element.
*
*
*/

package org.sablecc.sablecc;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import org.sablecc.sablecc.analysis.AnalysisAdapter;
import org.sablecc.sablecc.analysis.DepthFirstAdapter;
import org.sablecc.sablecc.node.*;

public class InternalTransformationsToGrammar extends DepthFirstAdapter
{

  private ResolveIds ids;
  private ResolveAltIds altIds;
  private ResolveTransformIds transformIds;
  private String currentProd;
  private int currentAlt;
  private String currentAltName;
  private boolean processingAst;
  private boolean processingProdTransform;

  static final int NONE = 0;
  static final int STAR = 1;
  static final int QMARK = 2;
  static final int PLUS = 3;

  int count;
  int elem;

  private List<Node> listSimpleTermTransform;

  public final Map<Node, String> simpleTermTransform;
  Map<String, List<PElem>> mapProductionTransformations;
  Map<Node, String> simpleTermOrsimpleListTermTypes;

  private Map<String, String> isElementIsAlist = new TreeMap<>();

  private List<PProd> listProd;

  public InternalTransformationsToGrammar(ResolveIds ids, ResolveAltIds altIds,
                                          ResolveTransformIds transformIds,
                                          List<Node> listSimpleTermTransform,
                                          Map<Node, String> simpleTermTransform,
                                          Map<String, List<PElem>> mapProductionTransformations,
                                          Map<Node, String> simpleTermOrsimpleListTermTypes)
  {
    this.ids = ids;
    this.altIds = altIds;
    this.transformIds = transformIds;
    this.listSimpleTermTransform = listSimpleTermTransform;
    this.simpleTermTransform = simpleTermTransform;
    this.mapProductionTransformations = mapProductionTransformations;
    this.simpleTermOrsimpleListTermTypes = simpleTermOrsimpleListTermTypes;
  }

  @Override
  public void inAProductions(AProductions node)
  {
    listProd = node.getProds();
  }

  private List<PAlt> listOfAlts;

  @Override
  public void inAAst(AAst node)
  {
    processingAst = true;
  }

  @Override
  public void outAAst(AAst node)
  {
    processingAst = false;
  }

  @Override
  public void caseAProd(AProd node)
  {
    currentProd = ids.names.get(node);
    listOfAlts = new LinkedList<>();

    for(PAlt alt : node.getAlts())
    {
      alt.apply(this);
    }

    node.setAlts(listOfAlts);
  }

  private List<PElem> listElems;
  private AAlt aParsedAlt;
  private List<String> listElemsAltTransform;
  private String currentNewAltName;

  boolean countElementNecessary;

  List<String> listOfAlternativeElemsWHaveName;

  private final Map<AElem, Integer> elemOperators = new HashMap<>();

  @Override
  public void caseAAlt(AAlt node)
  {
    count = 1;
    currentAltName = ids.names.get(node);

    AAltTransform currentAltTransform = (AAltTransform)node.getAltTransform();

    listOfAlternativeElemsWHaveName = new LinkedList<>();

    node.apply(new DepthFirstAdapter()
               {
                 @Override
                 public void inAElem(AElem node)
                 {
                   InternalTransformationsToGrammar.this.elemOperators.put(node, NONE);
                 }

                 @Override
                 public void caseAStarUnOp(AStarUnOp node)
                 {
                   count *= 2;
                   InternalTransformationsToGrammar.this.elemOperators.put((AElem)node.parent(), STAR);
                 }

                 @Override
                 public void caseAQMarkUnOp(AQMarkUnOp node)
                 {
                   count *= 2;
                   InternalTransformationsToGrammar.this.elemOperators.put((AElem)node.parent(), QMARK);
                 }

                 @Override
                 public void caseAPlusUnOp(APlusUnOp node)
                 {
                   InternalTransformationsToGrammar.this.elemOperators.put((AElem)node.parent(), PLUS);
                 }
               }
              );

    if(count == 1)
    {
      listElems = new LinkedList<>();
      listElemsAltTransform = new LinkedList<>();

      countElementNecessary = false;

      for(PElem pElem : node.getElems())
      {
        final PUnOp unOp = ((AElem)pElem).getUnOp();

        if( unOp != null &&
            ( unOp instanceof AQMarkUnOp ||
              unOp instanceof AStarUnOp )
          )
        {
          if(!countElementNecessary)
          {
            countElementNecessary = true;
          }
        }
      }

      for(PElem pElem : node.getElems())
      {
        pElem.apply(this);
      }

      TId nameOfAlt = null;

      if(node.getAltName() != null)
      {
        nameOfAlt = node.getAltName().clone();
      }

      currentNewAltName = currentProd + "." + currentAltName.toLowerCase();
      altIds.alts_elems.put(currentNewAltName, listElemsAltTransform);
      altIds.alts_elems_list_elemName.put(currentNewAltName, listOfAlternativeElemsWHaveName);

      AAltTransform altTransform = currentAltTransform.clone();

      AltTransformAdapter altTransformAdapter =
        new AltTransformAdapter(simpleTermTransform, listSimpleTermTransform,
                                currentNewAltName, altIds,
                                isElementIsAlist,
                                simpleTermOrsimpleListTermTypes);

      altTransform.apply(altTransformAdapter);

      aParsedAlt = new AAlt(nameOfAlt, listElems, altTransform);

      ids.names.put(aParsedAlt, ids.names.get(node));

      listOfAlts.add(aParsedAlt);
    }
    else
    {
      int max = count;
      AAltTransform altTransform;

      for(count = 0; count < max; count++)
      {
        listElems = new LinkedList<>();
        listElemsAltTransform = new LinkedList<>();

        elem = 0;

        currentNewAltName = currentProd + "." + currentAltName.toLowerCase()+(count + 1);

        countElementNecessary = false;

        for(PElem pElem : node.getElems())
        {
          final PUnOp unOp = ((AElem)pElem).getUnOp();

          if( unOp != null &&
              ( unOp instanceof AQMarkUnOp ||
                unOp instanceof AStarUnOp )
            )
          {
            if(!countElementNecessary)
            {
              countElementNecessary = true;
            }
          }
        }

        for(PElem pElem : node.getElems())
        {
          pElem.apply(this);
        }

        altIds.alts_elems.put(currentNewAltName, listElemsAltTransform);
        altIds.alts_elems_list_elemName.put(currentNewAltName, listOfAlternativeElemsWHaveName);

        altTransform = currentAltTransform.clone();

        AltTransformAdapter altTransformAdapter =
          new AltTransformAdapter(simpleTermTransform, listSimpleTermTransform,
                                  currentNewAltName, altIds,
                                  isElementIsAlist,
                                  simpleTermOrsimpleListTermTypes);

        altTransform.apply(altTransformAdapter);
        aParsedAlt = new AAlt(new TId(currentAltName.toLowerCase()+(count + 1)), listElems, altTransform);

        String currentAltInlining;
        currentAltInlining = "A" + ResolveIds.name(aParsedAlt.getAltName().getText()) + currentProd;
        ids.names.put(aParsedAlt, currentAltInlining);

        listOfAlts.add(aParsedAlt);
      }
    }
  }

  List<String> checkCreationOfXElem = new LinkedList<>();

  //It's also available for Ignored alternatives
  @Override
  public void caseAElem(AElem node)
  {
    if(!processingAst)
    {
      int op = this.elemOperators.get(node);
      String name = ids.elemTypes.get(node);
      String numero = (countElementNecessary == true ? ""+(count+1) : "" );
      String qMarkOrPlusElemType;
      String elemNameOfElem = null;

      TId aElemName = null;
      PSpecifier specifier = null;

      if(node.getElemName() != null)
      {
        elemNameOfElem = node.getElemName().getText();
        aElemName = new TId(elemNameOfElem);
      }

      if(node.getSpecifier() != null)
      {
        if(node.getSpecifier() instanceof ATokenSpecifier)
        {
          specifier = new ATokenSpecifier();
        }
        else
        {
          specifier = new AProductionSpecifier();
        }
      }

      AElem aElem = null;
      String elemName = node.getId().getText();
      boolean ok = false;
      boolean oklist = false;

      switch(op)
      {
      case NONE:
        {
          aElem = new AElem(aElemName, specifier, new TId(elemName), null);
          if(elemNameOfElem != null)
          {
            ids.names.put(aElem, ResolveIds.name(elemNameOfElem));
          }
          else
          {
            ids.names.put(aElem, ResolveIds.name(elemName));
          }
          ok = true;
        }
        break;
      case STAR:
        {

          if((count & (1 << elem)) != 0)
          {
            qMarkOrPlusElemType = ids.elemTypes.get(node);
            List<PElem> tmpProdTransform = mapProductionTransformations.get(qMarkOrPlusElemType);

            if(!checkCreationOfXElem.contains("$" + elemName))
            {
              checkCreationOfXElem.add("$" + elemName);
              listProd.add( createXelemProduction("$" + elemName, qMarkOrPlusElemType,
                                                  name, tmpProdTransform) );
            }

            elemName = "$" + elemName;
            aElem = new AElem(aElemName, new AProductionSpecifier(), new TId(elemName), null);

            if(elemNameOfElem != null)
            {
              ids.names.put(aElem, ResolveIds.name(elemNameOfElem));
            }
            else
            {
              ids.names.put(aElem, ResolveIds.name(elemName));
            }

            ok = true;
            oklist = true;
          }

          elem++;
        }
        break;
      case QMARK:
        {
          if((count & (1 << elem)) != 0)
          {
            aElem = new AElem(aElemName, specifier, new TId(elemName), null);

            if(elemNameOfElem != null)
            {
              ids.names.put(aElem, ResolveIds.name(elemNameOfElem));
            }
            else
            {
              ids.names.put(aElem, ResolveIds.name(elemName));
            }

            ok = true;
          }

          elem++;
        }
        break;
      case PLUS:
        {
          qMarkOrPlusElemType = ids.elemTypes.get(node);
          List<PElem> tmpProdTransform = mapProductionTransformations.get(qMarkOrPlusElemType);

          if(!checkCreationOfXElem.contains("$" + elemName))
          {
            checkCreationOfXElem.add("$" + elemName);
            listProd.add( createXelemProduction("$" + elemName, qMarkOrPlusElemType,
                                                name, tmpProdTransform) );
          }

          elemName = "$" + elemName;
          aElem = new AElem(aElemName, new AProductionSpecifier(), new TId(elemName), null);

          if(elemNameOfElem != null)
          {
            ids.names.put(aElem, ResolveIds.name(elemNameOfElem));
          }
          else
          {
            ids.names.put(aElem, ResolveIds.name(elemName));
          }

          ok = true;
          oklist = true;
        }
        break;
      }

      if(ok)
      {
        if(aElemName != null)
        {
          listElemsAltTransform.add(aElemName.getText());
          if(oklist)
          {
            if(elemNameOfElem != null)
            {
              listOfAlternativeElemsWHaveName.add(elemNameOfElem);
            }
            isElementIsAlist.put(currentProd+"."+currentAltName.toLowerCase()+numero+aElemName.getText(),
                                 node.getId().getText());
          }
        }
        else
        {
          listElemsAltTransform.add(elemName);
          if(oklist)
          {
            if(elemNameOfElem != null)
            {
              listOfAlternativeElemsWHaveName.add(elemNameOfElem);
            }
            isElementIsAlist.put(currentProd+"."+currentAltName.toLowerCase()+numero+node.getId().getText(),
                                 node.getId().getText());
          }
        }
      }

      if(aElem != null)
      {
        listElems.add(aElem);
      }
    }
  }

  /*
    This method creates the production for star(*) and plus(+) substitution in the grammar
    elem* -> $elem |
             elem
    This creates the production ::
                                   $elem                                   {-> elem* } 

                            = {nonTerminal} $elem elem       {-> [$elem.elem elem] }
                            | {terminal}    elem             {-> [elem] }
                            ;
  */
  public AProd createXelemProduction(final String name, final String elemTypeName,
                                     String XproductionName,
                                     List<PElem> nodeProdTransform)
  {
    final String rname = name.substring(1);
    List<PAlt> listOfAltsXelem = new LinkedList<>();

    if(nodeProdTransform != null)
    {
      nodeProdTransform = cloneList(nodeProdTransform);

      //Creation of the production transformation for Xelem
      //if the production transformation is introduced by the software
      if(nodeProdTransform.size() == 1)
      {
        AElem elem = (AElem)nodeProdTransform.get(0);
        if(elem.getUnOp() == null && elem.getId().getText().equals(rname))
        {
          List<PElem> elemsProdTransform = new LinkedList<>();
          elemsProdTransform.add( new AElem( null, new AProductionSpecifier(), new TId(rname), new AStarUnOp() ) );
          nodeProdTransform = elemsProdTransform;
        }
      }

    }
    //That means elem is token type
    else
    {
      List<PElem> elemsProdTransform = new LinkedList<>();
      elemsProdTransform.add( new AElem( null, new ATokenSpecifier(), new TId(rname), new AStarUnOp() ) );
      nodeProdTransform = elemsProdTransform;
    }

    final List<String> listProdTransformationOfXelem = new LinkedList<>();

    AElem []temp_listProdTransform = nodeProdTransform.toArray(new AElem[0]);
    for(int i=0; i<temp_listProdTransform.length; i++)
    {
      temp_listProdTransform[i].apply( new DepthFirstAdapter()
                                       {
                                         @Override
                                         public void caseAElem(AElem node)
                                         {
                                           //The production transformation needs to have a star operator.
                                           node.setUnOp(new AStarUnOp(new TStar()));
                                           if(node.getElemName() != null)
                                           {
                                             listProdTransformationOfXelem.add( node.getElemName().getText() );
                                           }
                                           else
                                           {
                                             listProdTransformationOfXelem.add( node.getId().getText() );
                                           }
                                         }
                                       }
                                     );
    }

    //creation of the first AltTransform node
    AElem[] prodTransformElems = nodeProdTransform.toArray(new AElem[0]);

    final List<PTerm> listTerms_first = new LinkedList<>();

    for(int i = 0; i < prodTransformElems.length; i++)
    {
      prodTransformElems[i].apply(new AnalysisAdapter()
                                  {
                                    @Override
                                    public void caseAElem(AElem node)
                                    {
                                      String tmpNodeName = ( (node.getElemName() == null) ? node.getId().getText() :
                                                             node.getElemName().getText() );
                                      List<PListTerm> listAListTerm_first = new LinkedList<>();

                                      if(elemTypeName.startsWith("T"))
                                      {
                                        listAListTerm_first.add(new ASimpleListTerm(new ATokenSpecifier(),
                                                                new TId(rname), null ));
                                      }
                                      else
                                      {
                                        listAListTerm_first.add(new ASimpleListTerm(new AProductionSpecifier(),
                                                                new TId(rname),new TId(tmpNodeName) ) );
                                      }
                                      listTerms_first.add( new AListTerm(new TLBkt(), listAListTerm_first) );
                                    }
                                  }
                                 );
    }

    AAltTransform aAltTransform = new AAltTransform(new TLBrace(), listTerms_first, new TRBrace());

    //create the first list of elems  of an alternative
    List<PElem> elems = new LinkedList<>();
    AElem aElemFirstTobeAdded;
    //the elem is a token
    if(elemTypeName.startsWith("T"))
    {
      aElemFirstTobeAdded = new AElem(null, new ATokenSpecifier(), new TId(rname), null);
    }
    else
    {
      aElemFirstTobeAdded = new AElem(null, new AProductionSpecifier(), new TId(rname), null);
    }
    elems.add(aElemFirstTobeAdded);

    //creation of the first alternative
    AAlt aParsedAlt = new AAlt(new TId("terminal"), elems, aAltTransform);

    listOfAltsXelem.add(aParsedAlt);

    //create the second AltTransform node
    prodTransformElems = nodeProdTransform.toArray(new AElem[0]);

    final List<PTerm> listTerms_second = new LinkedList<>();

    for(int i = 0; i < prodTransformElems.length; i++)
    {
      prodTransformElems[i].apply(new AnalysisAdapter()
                                  {
                                    @Override
                                    public void caseAElem(AElem node)
                                    {
                                      String tmpNodeName = ( (node.getElemName() == null) ? node.getId().getText() :
                                                             node.getElemName().getText() );

                                      List<PListTerm> listAListTerm_second = new LinkedList<>();

                                      listAListTerm_second.add(new ASimpleListTerm(null, new TId(name),
                                                               new TId(tmpNodeName)) );

                                      if(elemTypeName.startsWith("T"))
                                      {
                                        listAListTerm_second.add(new ASimpleListTerm(new ATokenSpecifier(),
                                                                 new TId(rname), null ));
                                      }
                                      else
                                      {
                                        listAListTerm_second.add(new ASimpleListTerm(new AProductionSpecifier(),
                                                                 new TId(rname),
                                                                 new TId(tmpNodeName) ) );
                                      }
                                      listTerms_second.add(new AListTerm(new TLBkt(), listAListTerm_second));
                                    }
                                  }
                                 );
    }

    aAltTransform = new AAltTransform(new TLBrace(), listTerms_second, new TRBrace());

    //creation of the second list of elems of an alternative :: two elems
    elems = new LinkedList<>();

    //first elem
    AElem aElemSecondTobeAdded = new AElem(null, new AProductionSpecifier(), new TId(name), null);
    elems.add(aElemSecondTobeAdded);

    //second elem
    if(elemTypeName.startsWith("T"))
    {
      aElemSecondTobeAdded = new AElem(null, new ATokenSpecifier(), new TId(rname), null);
    }
    else
    {
      aElemSecondTobeAdded = new AElem(null, new AProductionSpecifier(),  new TId(rname), null);
    }
    elems.add(aElemSecondTobeAdded);

    aParsedAlt = new AAlt(new TId("non_terminal"), elems, aAltTransform);

    listOfAltsXelem.add(aParsedAlt);

    AProd prodToReturn = new AProd(new TId(name), new TArrow(), nodeProdTransform, listOfAltsXelem);
    prodToReturn.apply(ids);
    prodToReturn.apply(transformIds.getProdTransformIds());

    return prodToReturn;
  }

  @SuppressWarnings("unchecked")
  private <T extends Node> List<T> cloneList(List<T> list)
  {
    List<T> clone = new LinkedList<>();

    for (final T node : list) {
      clone.add((T)node.clone());
    }
    return clone;
  }

  private String xproductionType(String name)
  {
    return "P$" + name.substring(1).toLowerCase();
  }

}
