/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * This file is part of SableCC.                             *
 * See the file "LICENSE" for copyright information and the  *
 * terms and conditions for copying, distribution and        *
 * modification of SableCC.                                  *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

package org.sablecc.sablecc;

import java.util.Iterator;
import java.util.Map;
import java.util.TreeMap;

final class LR0ItemSet implements Cloneable, Comparable<LR0ItemSet>
{
  private final Map<LR0Item, LR0Item> items;
  private int hashCode;

  LR0ItemSet()
  {
    items = new TreeMap<>();
  }

  private LR0ItemSet(LR0ItemSet set
                      )
  {
    items = new TreeMap<>(set.items);
  }

  void set
    (LR0Item item)
  {
    if(items.put(item, item) == null)
    {
      hashCode += item.hashCode();
      modified_ = true;
    }
  }

  boolean get
    (LR0Item item)
  {
    return items.get(item) != null;
  }

  LR0Item[] items_;
  boolean modified_ = true;

  private void computeArray()
  {
    items_ = items.keySet().toArray(new LR0Item[0]);
    modified_ = false;
  }

  LR0Item[] items()
  {
    if(modified_)
    {
      computeArray();
    }

    return items_;
  }

  @Override
  public String toString()
  {
    StringBuilder result = new StringBuilder();
    result.append("{");

    Production[] productions = Production.productions();
    boolean space = false;
    for(int i = 0; i < productions.length; i++)
    {
      int rightsideLength = productions[i].rightside().length;

      for(int j = 0; j <= rightsideLength; j++)
      {
        LR0Item item = new LR0Item(productions[i].index, j);
        if(get
            (item))
        {
          if(space)
          {
            result.append(",");
          }
          else
          {
            space = true;
          }

          result.append(item);
        }
      }

    }

    result.append("}");
    return result.toString();
  }

  @Override
  public LR0ItemSet clone()
  {
    return new LR0ItemSet(this);
  }

  @Override
  public boolean equals(Object obj)
  {
    if((obj == null) ||
        (obj.getClass() != this.getClass()))
    {
      return false;
    }

    LR0ItemSet set
      = (LR0ItemSet) obj;

    return items.keySet().equals(set.items.keySet());
  }

  @Override
  public int hashCode()
  {
    return hashCode;
  }

  @Override
  public int compareTo(LR0ItemSet set)
  {
    int result = items.size() - set.items.size();

    if(result == 0)
    {
      Iterator<LR0Item> e = items.keySet().iterator();
      Iterator<LR0Item> f = set.items.keySet().iterator();

      while(e.hasNext() && f.hasNext() && (result == 0))
      {
        result = e.next().compareTo(f.next());
      }

      if(result == 0)
      {
        if(e.hasNext())
        {
          return 1;
        }

        if(f.hasNext())
        {
          return -1;
        }
      }
    }

    return result;
  }

}
